/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Record.java 511426 2007-02-25 03:25:02Z vgritsenko $
 */

package org.apache.xindice.core.data;

import java.util.Map;

/**
 * Record is a container for Key/Value pair.  Record also provides
 * metadata for a Value stored in the database.
 *
 * @version $Revision: 511426 $, $Date: 2007-02-24 22:25:02 -0500 (Sat, 24 Feb 2007) $
 */
public final class Record {
    public static final String CREATED = "created";
    public static final String MODIFIED = "modified";
    public static final String OWNER = "owner";
    public static final String GROUP = "group";

    private final Key key;
    private Value value;
    private final Map meta;


    public Record() {
        this.key = null;
        this.meta = null;
    }

    public Record(Key key, Value value, Map meta) {
        this.key = key;
        this.value = value;
        this.meta = meta;
    }

    public Record(Key key, Value value) {
        this.key = key;
        this.value = value;
        this.meta = null;
    }

    /**
     * getKey returns the Record's Key.
     *
     * @return The Record's Key
     */
    public Key getKey() {
        return key;
    }

    /**
     * setValue sets the Record's Value.
     *
     * @param value The new Value
     */
    public void setValue(Value value) {
        this.value = value;
    }

    /**
     * setValue sets the Record's Value.
     *
     * @param value The new Value
     */
    public void setValue(String value) {
        this.value = new Value(value);
    }

    /**
     * getValue returns the Record's Value.
     *
     * @return The Record's Value
     */
    public Value getValue() {
        return value;
    }

    /**
     * getMetaData returns metadata about the Value.
     *
     * @param name The property name
     * @return The metadata value
     */
    public Object getMetaData(Object name) {
        return meta != null ? meta.get(name) : null;
    }
}
